#include "../config.h"

#include <unistd.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include "inotify.h"
#include "inotify-syscalls.h"
#include "inotifyutils.h"

#define _GNU_SOURCE
#include <getopt.h>

extern char *optarg;
extern int optind, opterr, optopt;

typedef struct {
	int wd;
	char * filename;
} watch;

#define EXIT_OK 0
#define EXIT_ERROR 1
#define EXIT_TIMEOUT 2



// METHODS
bool parse_opts(
  int * argc,
  char *** argv,
  int * events,
  bool * monitor,
  int * quiet,
  unsigned long int * timeout,
  bool * filename
);

void print_help();

watch * watch_files( int numfiles, char ** filenames, int events, int inotify_fd );

watch * find_watch( watch * watches, int num_watches, int wd );


int main(int argc, char ** argv)
{
	int events = 0;
	bool monitor = false;
	int quiet = 0;
	unsigned long int timeout = 0;
	bool filename = false;

	// Parse commandline options, aborting if something goes wrong
	if ( !parse_opts(&argc, &argv, &events, &monitor, &quiet, &timeout, &filename) )
		return EXIT_ERROR;


	// Try to initialise inotify
	int inotify_fd = inotify_init();
	if (inotify_fd < 0)
	{
		fprintf(stderr, "inotify_init() failed.  Are you running Linux 2.6.13 "
		                "or later?  If so, something mysterious has gone "
		                "wrong.  Please e-mail " PACKAGE_BUGREPORT " and "
		                "mention that you saw this message.\n");
		return EXIT_ERROR;
	}


	// Attempt to watch file
	// If events is still 0, make it all events.
	if (events == 0)
		events = IN_ALL_EVENTS;

	watch * watches = watch_files( argc, argv, events, inotify_fd );

	// Now wait till we get event
	static const int MAX_EVENTS = 4096;
	struct inotify_event event[MAX_EVENTS];
	event[0].wd = 0;
	event[0].mask = 0;
	event[0].cookie = 0;
	event[0].len = 0;
	ssize_t bytes = 0;

	fd_set read_fds;
	struct timeval read_timeout;
	read_timeout.tv_sec = timeout;
	read_timeout.tv_usec = 0;
	struct timeval * read_timeout_ptr = ( timeout <= 0 ? NULL : &read_timeout );

	do {
		bytes = 0;
		while ( bytes < sizeof(struct inotify_event) ) {
			FD_ZERO(&read_fds);
			FD_SET(inotify_fd, &read_fds);
			int rc = select(inotify_fd + 1, &read_fds,
			                NULL, NULL, read_timeout_ptr);
			if ( rc < 0 ) {
				perror("Select on inotify failed");
				return EXIT_ERROR;
			}
			else if ( rc == 0 )
			{
				return EXIT_TIMEOUT;
			}
			int thisBytes = read(inotify_fd, &event + bytes, sizeof(struct inotify_event)*MAX_EVENTS - bytes);
			if ( thisBytes < 0 ) {
				perror("Read from inotify failed");
				return EXIT_ERROR;
			}
			if ( thisBytes == 0 ) {
				fprintf(stderr, "Inotify reported end-of-file.  Possibly too many events occurred at once.\n");
				return EXIT_ERROR;
			}
			bytes += thisBytes;
		}

		watch * this_watch = find_watch( watches, argc, event[0].wd );
		if ( !this_watch ) {
			fprintf(stderr, "Inotify returned event on invalid wd %d\n", event[0].wd);
			continue;
		}

		// Print out the event(s)
		static struct inotify_event * this_event;
		static char * this_event_char;
		this_event_char = (char *)&event[0];
		static int remaining_bytes;
		remaining_bytes = bytes;
		do {
			this_event = (struct inotify_event *)this_event_char;

			if ( quiet < 2 ) {
				static char * this_event_str;
				if ( filename )
					printf("%s, ", this_watch->filename);
				this_event_str = event_to_str(this_event->mask);
				printf("%s", this_event_str);
				free(this_event_str);
				if ( (this_event->len > 0) && (quiet < 1) )
					printf(" %s", this_event->name );
				printf("\n");
			}

			this_event_char += sizeof(struct inotify_event) + this_event->len;
			remaining_bytes -= sizeof(struct inotify_event) + this_event->len;
		} while ( monitor && remaining_bytes >= sizeof(struct inotify_event) );

		// I _think_ this should never happen.
		if ( monitor && remaining_bytes != 0 ) {
			fprintf(
			  stderr,
			  "%f event(s) may have been lost!\n",
			  ((float)remaining_bytes)/((float)sizeof(struct inotify_event))
			);
		}

		fflush( NULL );

	} while ( monitor );

	// If we weren't trying to listen for this event...
	if ( (events & event[0].mask) == 0 )
	{
		// ...then most likely something bad happened, like IGNORE etc.
		return EXIT_ERROR;
	}

	return EXIT_OK;
}




bool parse_opts(
  int * argc,
  char *** argv,
  int * events,
  bool * monitor,
  int * quiet,
  unsigned long int * timeout,
  bool * filename
) {
	assert( argc ); assert( argv ); assert( events ); assert( monitor );
	assert( quiet ); assert( timeout ); assert( filename );

	// Short options
	char * opt_string = "mhqt:fe:";

	// Construct array
	struct option long_opts[7];

	// --help
	long_opts[0].name = "help";
	long_opts[0].has_arg = 0;
	long_opts[0].flag = NULL;
	long_opts[0].val = (int)'h';
	// --event
	long_opts[1].name = "event";
	long_opts[1].has_arg = 1;
	long_opts[1].flag = NULL;
	long_opts[1].val = (int)'e';
	int new_event;
	// --monitor
	long_opts[2].name = "monitor";
	long_opts[2].has_arg = 0;
	long_opts[2].flag = NULL;
	long_opts[2].val = (int)'m';
	// --quiet
	long_opts[3].name = "quiet";
	long_opts[3].has_arg = 0;
	long_opts[3].flag = NULL;
	long_opts[3].val = (int)'q';
	// --timeout
	long_opts[4].name = "timeout";
	long_opts[4].has_arg = 1;
	long_opts[4].flag = NULL;
	long_opts[4].val = (int)'t';
	char * timeout_end = NULL;
	// --filename
	long_opts[5].name = "f";
	long_opts[5].has_arg = 0;
	long_opts[5].flag = NULL;
	long_opts[5].val = (int)'f';
	// Empty last element
	long_opts[6].name = 0;
	long_opts[6].has_arg = 0;
	long_opts[6].flag = 0;
	long_opts[6].val = 0;

	// Get first option
	char curr_opt = getopt_long(*argc, *argv, opt_string, long_opts, NULL);

	// While more options exist...
	while ( (curr_opt != '?') && (curr_opt != (char)-1) )
	{
		switch ( curr_opt )
		{
			// --help or -h
			case 'h':
				print_help();
				// Shouldn't process any further...
				return false;
				break;

			// --monitor or -m
			case 'm':
				*monitor = true;
				break;

			// --quiet or -q
			case 'q':
				(*quiet)++;
				break;

			// --filename or -f
			case 'f':
				*filename = true;
				break;

			// --timeout or -t
			case 't':
				*timeout = strtoul(optarg, &timeout_end, 10);
				if ( *timeout_end != '\0' || *timeout < 0)
				{
					fprintf(stderr, "'%s' is not a valid timeout value.\n"
					        "Please specify an integer of value 0 or greater.\n",
					        optarg);
					return false;
				}
				break;

			// --event or -e
			case 'e':
				// Get event mask from event string
				new_event = str_to_event(optarg);

				// If optarg was invalid, abort
				if ( new_event == -1 )
				{
					fprintf(stderr, "'%s' is not a valid event!  Run with the "
					                "'--help' option to see a list of "
					                "events.\n", optarg);
					return false;
				}

				// Add the new event to the event mask
				(*events) = ( (*events) | new_event );

				break;


		}

		curr_opt = getopt_long(*argc, *argv, opt_string, long_opts, NULL);

	}

	if ( *monitor && *timeout != 0 )
	{
		fprintf(stderr, "-m and -t cannot both be specified.\n");
		return false;
	}

	(*argc) -= optind;
	*argv = &(*argv)[optind];


	// Make sure we were passed at least one file
	if ( (*argc) < 1 )
	{
		fprintf(stderr, "No files specified to watch!\n");
		return false;
	}


	// If ? returned, invalid option
	return (curr_opt != '?');
}


watch * watch_files( int numfiles, char ** filenames, int events, int inotify_fd ) {
	assert( filenames );

	watch * watches = (watch *)malloc( sizeof(watch)*numfiles );
	assert( watches );

	for ( int i = 0; i < numfiles; i++ ) {
		static int wd;
		wd = inotify_add_watch (inotify_fd, filenames[i], events);
		if (wd < 0) {
			if ( wd == -1 ) {
				fprintf(stderr, "Failed to watch %s: %s\n", filenames[i], strerror(-wd));
				fprintf(stderr,"Possibly file doesn't exist, or you don't have "
				               "sufficient permissions to watch\nit.\n");
				free( watches );
				return NULL;
			} // if ( wd == -1 )
			else {
				fprintf(
				  stderr,
				  "Failed to watch %s: returned wd was %d (expected -1 or >0 )",
				  filenames[i],
				  wd
				);
				free( watches );
				return NULL;
			} // else
		} // if ( wd < 0 )
		watches[i].wd = wd;
		watches[i].filename = filenames[i];
	} // for

	return watches;
}



void print_help()
{
	printf("inotifywait %s\n", PACKAGE_VERSION);
	printf("By Rohan McGovern, %s\n", PACKAGE_BUGREPORT);
	printf("Usage: inotifywait [ options ] file1 [ file2 ] [ file3 ] [ ... ]\n");
	printf("Options:\n");
	printf("\t-h|--help    \tShow this help text.\n");
	printf("\t-m|--monitor \tKeep listening for events forever.  Without\n"
	       "\t             \tthis option, inotifywait will exit after one\n"
	       "\t             \tevent is received.\n");
	printf("\t-q|--quiet   \tWhen listening on a directory, do not print the\n"
	       "\t             \tfilename(s) which events refer to.\n");
	printf("\t-qq          \tDo not print anything at all.\n");
	printf("\t-t|--timeout <seconds>\n"
	       "\t             \tWhen listening for a single event, time out after\n"
	       "\t             \twaiting for an event for <seconds> seconds.\n"
	       "\t             \tIf <seconds> is 0, inotifywait will never time out.\n");
	printf("\t-f|--filename\tFor each event, print the filename on which the\n"
	       "\t             \tevent occurred.\n");
	printf("\t-e|--event <event1> [ -e|--event <event2> ... ]\n"
	       "\t\tListen for specific event(s).  If omitted, all events are \n"
	       "\t\tlistened for.\n\n");
	printf("Exit status:\n");
	printf("\t%d  -  An event you asked to watch for was received.\n", EXIT_OK );
	printf("\t%d  -  An event you did not ask to watch for was received\n", EXIT_ERROR);
	printf("\t      (usually delete_self or unmount), or some error occurred.\n");
	printf("\t%d  -  The --timeout option was given and no events occurred\n", EXIT_TIMEOUT);
	printf("\t      in the specified interval of time.\n\n");
	printf("Events:\n");
	printf("\taccess\t\tfile was accessed\n");
	printf("\tmodify\t\tfile was modified\n");
	printf("\tattrib\t\tfile attributes changed\n");
	printf("\tclose_write\tfile closed, after being opened in writeable "
	       "mode\n");
	printf("\tclose_nowrite\tfile closed, after being opened in read-only "
	       "mode\n");
	printf("\tclose\t\tfile closed, regardless of read/write mode\n");
	printf("\topen\t\tfile was opened\n");
	printf("\tmove\t\ta file within watched directory was moved\n");
	printf("\tcreate\t\ta file was created within watched directory\n");
	printf("\tdelete\t\ta file was deleted within watched directory\n");
	printf("\tdelete_self\tthe watched file was deleted\n");
	printf("\tunmount\t\tfile system on which watched file is located was\n"
	       "\t       \t\tunmounted\n");
}

watch * find_watch( watch * watches, int num_watches, int wd ) {
	assert( watches );

	watch * found_watch = NULL;
	for ( int i = 0; i < num_watches && !found_watch; ++i ) {
		if ( watches[i].wd == wd )
			found_watch = &watches[i];
	}
	return found_watch;
}
